/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2016. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function() {

dojo.provide("com.ibm.rdm.web.editor.editorutils");
dojo.require("jazz.util.html");
dojo.require("com.ibm.rdm.web.validation.ValidationServiceClient");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.util.queryutils");
dojo.require("com.ibm.rdm.web.common.ErrorHandler");
dojo.require("com.ibm.rdm.web.editor.ElementIdPolicy");
dojo.require("com.ibm.rdm.web.editor.EditorConstants");
dojo.require("com.ibm.rdm.web.editor.RMCleanDialog");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.MessageForm");
dojo.require("com.ibm.rdm.web.common.ResourceLinksHandler");
dojo.require("com.ibm.rdm.web.viewer.EmbeddedResourceLoader");
dojo.require("jazz.ui.MessageBox");

dojo.requireLocalization("com.ibm.rdm.web.editor", "RichTextEditor");

var ValidationServiceClient = com.ibm.rdm.web.validation.ValidationServiceClient;
var ErrorHandler = com.ibm.rdm.web.common.ErrorHandler;
var MessageHandler = com.ibm.rdm.web.common.MessageHandler;
var MessageForm = com.ibm.rdm.web.common.MessageForm;
var MessageBox = jazz.ui.MessageBox;
var editorUtilMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.editor", "RichTextEditor");
var EditorConstants = com.ibm.rdm.web.editor.EditorConstants;
var ResourceLinksHandler = com.ibm.rdm.web.common.ResourceLinksHandler;
var domutils = com.ibm.rdm.web.util.domutils;
var EmbeddedResourceLoader = com.ibm.rdm.web.viewer.EmbeddedResourceLoader;

// Regular expressions to detect characters that are considered to be word characters or non-word characters during
// automatic word selection in glossary term lookup.
// We cannot use simple \w and \W selectors, since they only consider ASCII characters [0-9,a-z,A-Z], so things
// like accented characters are then considered non-word characters.
// We cannot use simple \s and \S selectors, since they would lead to common puctuation characters being considered
// to be inside words and the selection being too greedy when the word is preceeded or followed by some
// common punctuation character like (.;:![])
var REGEX_WORD_CHARACTERS = /[^\s,\u0021-\u002f,\u003a-\u0040,\u005b-\u0060,\u007b-\u007f]/;
var REGEX_NON_WORD_CHARACTERS = /[\s,\u0021-\u002f,\u003a-\u0040,\u005b-\u0060,\u007b-\u007f]/;

com.ibm.rdm.web.editor.editorutils = {
	
	prepareHtmlForServer: function(/*String|Node*/html, skipNamespaceWrapping, assignCommentIds) {
		// summary:
		//		Cleanup spurious data that CKEditor adds to rich text source.  Also work around
		//		SAXParser issue with &nbsp.  See comments in the code.
		// description:
		//		This method cleans up spurious data that CKEditor adds to the html source, works around
		//		the SAXParser issue with &nbsp, and creates well-formed xhtml.  For more information,
		//		see the comments in the code.
		// returns:
		//		<String> html text properly formed and with unnecessary tags removed
		
		var div;
		
		if (typeof html === "string") {
			div = document.createElement("div");
			div.innerHTML = html;
		} else {
			div = html;
		}

		// workaround 32361, which is a FF bug, only do for rich text documents, as comments store
		// urls relatively
		var i,
			url,
			urlString,
			images,
			imageNode,
			imgNode,
			imageSrc,
			revIndex;
			
		if (dojo.isFF && assignCommentIds) {
			// domain name is stripped from absolute urls when domain name is the same as the server,
			// fix this so that urls are saved properly
			var links = dojo.query("a", div);
			for (i = 0; i < links.length; i++) {
				var link = links[i];
				var linkHref = link.attributes.getNamedItem("href");
				if (linkHref) {
					if (linkHref.value.indexOf("http") != 0) {
						url = new dojo._Url(com.ibm.rdm.web.FRONTING_SERVICE_RESOURCES, linkHref.value);
						urlString = url.toString();
						link.setAttribute("href", urlString);
					}
				}
			}
			images = dojo.query("img", div);
			for (i = 0; i < images.length; i++) {
				var image = images[i];
				var imgSrc = image.attributes.getNamedItem("src").value;
				if (imgSrc) {
					if (imgSrc.indexOf("http") != 0) {
						url = new dojo._Url(com.ibm.rdm.web.FRONTING_SERVICE_RESOURCES, imgSrc);
						urlString = url.toString();
						image.setAttribute("src", urlString);
					}
				}
			}
		}
					
		// must remove embedded contents if there are any embeds in the content
				
		var clonedNode = false;
		if (assignCommentIds) {
			// we're validating content for the resource, not a comment
			var embeds = [];
			embeds = dojo.query(EditorConstants.RICH_TEXT_EMBED_CONTENT_SELECTOR, div);
			var imageErrors = dojo.query(EditorConstants.IMG_EMBED_EDIT_ERROR_CONTENT, div);
			images = dojo.query("img", div);
			if (embeds.length > 0 || imageErrors.length > 0 || images.length > 0) {
				if (typeof html === "string") {
					embeds.forEach(function(node) {
						node.parentNode.removeChild(node);
					});
					imageErrors.forEach(function(node) {
						dojo.removeAttr(node, "class");
					});
					// We use the revision of an image when displaying embedded images to get around
					// browser caching issues with images.  The revision should not be saved though,
					// so remove it.
					images = dojo.query("img", div);
					for (i = 0; i < images.length; i++) {
						imageNode = images[i];
						imageSrc = imageNode.src;
						revIndex = imageSrc.indexOf("&revision=");
						if (revIndex != -1) {
							imageNode.src = imageSrc.substring(0, revIndex);
						}
					}
				} else {
					// actual editor node is passed to this function, so create a new div
					var newDiv = div.cloneNode(true);
					clonedNode = true;
					embeds = dojo.query(EditorConstants.RICH_TEXT_EMBED_CONTENT_SELECTOR, newDiv);
					embeds.forEach(function(node) {
						node.parentNode.removeChild(node);
					});
			
					// fix up embedded image errors which are modified when in edit mode
					imageErrors = dojo.query(EditorConstants.IMG_EMBED_EDIT_ERROR_CONTENT, newDiv);
					for (i = 0; i < imageErrors.length; i++) {
						imgNode = imageErrors[i];
						dojo.removeAttr(imgNode, "class");
					}
					// We use the revision of an image when displaying embedded images to get around
					// browser caching issues with images.  The revision should not be saved though,
					// so remove it.
					images = dojo.query("img", newDiv);
					for (i = 0; i < images.length; i++) {
						imageNode = images[i];
						imageSrc = imageNode.src;
						revIndex = imageSrc.indexOf("&revision=");
						if (revIndex != -1) {
							imageNode.src = imageSrc.substring(0, revIndex);
						}
					}
					div = newDiv;
				}
			}
		}
		
		// Paragraphs containing embeds were changed to <div> for the editor.  Change them back again.
		dojo.query("."+EditorConstants.REPLACED_EMBED_PARAGRAPH_ELEMENT_CLASS, div).forEach(function(replacedParagraph){
			var replacementNode = null;
			// work item 74625: putting embedded artifact in a list creates bad xhtml,
			// only change the node if it is actually a div
			if (replacedParagraph.nodeName.toLowerCase() === "div") {
				replacementNode = domutils.changeNodeTag(replacedParagraph, "p");
			} else {
				replacementNode = replacedParagraph;
			}
			dojo.removeClass(replacementNode, EditorConstants.REPLACED_EMBED_PARAGRAPH_ELEMENT_CLASS);
			// make sure class attribute is removed if no longer required.
			if (dojo.hasAttr(replacementNode, "class")) {
				// empty class value
				dojo.removeAttr(replacementNode, "class");
			}
		});
		
		// Assign unique comment IDs to the appropriate elements.
		var assignIds = com.ibm.rdm.web.editor.ElementIdPolicy.getDefault();
			
		assignIds({
			rootElement: div,
			assignCommentIds: assignCommentIds
		});
			
		// workaround for work item 43240, font element is deprecated, so
		// server validation fails
		dojo.query("font", div).forEach(function (fontElement){
			var parent = fontElement.parentNode;
			var childNodes = fontElement.childNodes;
			for (var i=0; i<childNodes.length; i++) {
				parent.appendChild(childNodes[i]);
			}
			parent.removeChild(fontElement);
		});
		
		dojo.query("table", div).forEach(function (tableElement){
			// workaround work item 65754
			if (dojo.isFF > 10) {
				var heightAttr = dojo.attr(tableElement, "height");
				if (heightAttr) {
					dojo.style(tableElement, "height", heightAttr + "px");
				}
			}
			// workaround for 79647
			var borderAttr = dojo.attr(tableElement, "border");
			if (borderAttr) {
				dojo.attr(tableElement, "border", Math.round(borderAttr));
			}
		});
		
		dojo.query("ol, ul", div).forEach(function (list) {
			// workaround for 90523: "Error saving text artefact containing bulleted list"
			// remove ol or ul list if it has no list items, preventing an error message
			var listItems = list.getElementsByTagName('li');
			if (listItems.length === 0) {
				list.parentNode.removeChild(list);
			}
		});
		
		// Remove the temporary class added to the embedded artifact
		var embeddedArtifactSpaces = dojo.query("span." + EditorConstants.EMBEDDED_CONTENT_SPACE, div);
		for (i = 0; i < embeddedArtifactSpaces.length; i++) {
			spaceNode = embeddedArtifactSpaces[i];
			dojo.removeAttr(spaceNode, "class");
		}
		
		var illegalAttrs =  ["dojoattachpoint", "data-cke-saved-src", "data-cke-eol", "data-cke-saved-href",
									"_zipIdx", "_zipidx", "contentEditable", "unselectable", "_fck_bookmark",
									"data-cke-bookmark", "data-cke-expando", "data-cke-temp", "data-cke-editable",
									"data-cke-saved-name", "cke-saved-name", "data-cke-highlight", "data-cke-realelement",
									"data-cke-resizable", "data-cke-real-node-type", "data-cke-real-element-type",
									"data-nostyle", "data-no-magicline", "data-pasted-embed", "firebugversion", "size", "height"];
		
		// Format this array in a query friendly string that will be used to find offending nodes
		var illegalAttrsQueryString = illegalAttrs.map(function(attr){
			return '['+attr+']';
		}).join(', ');
	
		var removeIllegalAttrs = function(n) {
			dojo.forEach(illegalAttrs, function(attr){
				if (dojo.hasAttr(n, attr)) {
					dojo.removeAttr(n, attr);
				}
			});
		};
	var clean = dojo.hitch(this, function (n) {
		if (dojo.hasClass(n, EditorConstants.JAZZ_RESOURCE_LINK_CLASS)) {
			this.cleanupJazzResourceLink(n);
		}
		if (dojo.hasAttr(n, "itemId")) {
			/* RTC Rich client - copy URL from title menu */
			dojo.removeAttr(n, "itemId");
			// IE gets font tags from RTC rich client, not allowed...
			dojo.query("font", n).forEach(function (fontTag){
				var parent = fontTag.parentNode;
				parent.appendChild(fontTag.firstChild);
				parent.removeChild(fontTag);
			});
		}
		if (n.nodeName.toLowerCase() == "tr") {
			// workaround 31959, remove empty table rows (rows with no
			// columns)
			if (n.childNodes.length == 0) {
				n.parentNode.removeChild(n);
			}
		}
		removeIllegalAttrs(n);
	});
	
	// find the offending attributes in one query, this takes an almost unmeasurable amount
	// of time on modern browsers
	var nodes = dojo.query("div[firebugversion], hr[size], table[height], a[itemId], a.jazz-ui-ResourceLink, tr, img[data-cke-saved-src], a[data-cke-saved-href], " + illegalAttrsQueryString, div);

	nodes.forEach(clean);

	// work item 79528: Document rendering in view mode - scrollbar messed up
	dojo.query(".dijitContentPane", div).removeClass("dijitContentPane");
			
		// TODO - &nbsp causes server parsing error, replace all &nbsp tags
		var contents = div.innerHTML;
		contents = this.removeNonBreakingSpaceEntities(contents);

		// TODO - create valid html, on IE after assigning text to
		// innerHtml of div, table data is no longer well-formed and causes
		// SAX parser errors
		var writer = new CKEDITOR.htmlWriter();
		writer.lineBreakChars = '';
		var fragment = CKEDITOR.htmlParser.fragment.fromHtml(contents);
		fragment.writeHtml(writer);
		contents = writer.getHtml();

		if (!skipNamespaceWrapping) {
			contents = domutils.wrapForXHTMLValidation(contents);
		}
		if ((typeof html === "string") || clonedNode) {
			// destroy the div we created or cloned, it hasn't been appended to the DOM, so
			// we need to explicitly destroy it
			dojo.destroy(div);
		}
		return contents;
				
	},
	
	cleanupJazzResourceLink: function(linkNode) {
		// summary:
		//		When a jazz resource link is pasted into text, we must clean it up to remove
		//		decorations and to remove invalid attributes.
		
		dojo.query("img.arrow", linkNode).forEach(function (img){
			img.parentNode.removeChild(img);
		});
			
		dojo.removeAttr(linkNode, "class");
		dojo.removeAttr(linkNode, "widgetId");
		dojo.removeAttr(linkNode, "tabIndex");
		dojo.removeAttr(linkNode, "target");
	},

	handleCommonOnPaste: function(evt) {
		// summary:
		//		Perform common processing for pasted text.  The processing should apply to both
		//		comments and rich text documents.
		// evt:
		//		Paste event passed from CKEditor
		// returns:
		//		Modified html or null if pasted data is in text format
		var pasteData = evt.data;
		// if paste as text, do nothing
		if (pasteData.type && pasteData.type !== "html") {
			return null;
		}
		var pasteHtml = pasteData.dataValue;
		
		// In Chrome, the editor sometimes added an extra &nbsp; in front of <span> &nbsp;</span>. We need to remove that extra &nbsp; in
		// the embedded artifact, otherwise it will show extra space in reporting.
		if (dojo.isChrome) {
			var targetStr = '<span class="' + EditorConstants.EMBEDDED_CONTENT_SPACE;
			var searchStr = '&nbsp;' + targetStr;
			pasteHtml = pasteHtml.replace(new RegExp(searchStr, 'g'), targetStr);
		}
		
		var pasteDiv = document.createElement("div");
		pasteDiv.innerHTML = pasteHtml;
		
		dojo.query(EditorConstants.COMMENTABLE_ELEMENTS, pasteDiv).forEach(function (node){
			// remove highlight outline if copy occurred from requirement editor
			// in view mode and an element is selected
			if (dojo.hasClass(node, EditorConstants.RICH_TEXT_NODE_SELECTED)) {
				dojo.removeClass(node, EditorConstants.RICH_TEXT_NODE_SELECTED);
			}
		});
			
		// remove images that reference external resources
		var images = dojo.query("img", pasteDiv);
		images.forEach(function(node) {
			if (!domutils.isBase64Image(node)) {
				if (!ResourceLinksHandler.isInternalWrappedResource(node.src) && !ResourceLinksHandler.isLinkInternal(node.src)) {
					node.parentNode.removeChild(node);
				}
			}
		});
		
		// Resolve <a> tags of the form:
		//		https://<server>:<port>/rm/web#action=com.ibm.rdm.web.pages.showArtifact&artifactURI=<path to artifactURI>
		// to be a basic resource uri
		var links = dojo.query("a", pasteDiv);
		links.forEach(function(node) {
			var target = node.getAttribute("href");
			target = domutils.getXMLDecodedText(target);
			var resolvedTarget = ResourceLinksHandler.getArtifactUriFromExternalLink(target);
			if (resolvedTarget) {
				node.setAttribute("href", resolvedTarget);
			}
		});
		
		// Fix for defect 104784
		var titleSpans = dojo.query(".resource-id, .resource-title, .resource-title-node", pasteDiv);
		titleSpans.forEach(function(node) {
			dojo.attr(node, "style", "");
			dojo.attr(node, "class", "");
		});
		
		pasteHtml = pasteDiv.innerHTML;

		if (pasteHtml === "") {
			pasteHtml = "<p></p>";
		}
		dojo.destroy(pasteDiv);
		
		// work item 35686 - replace consecutive <br> tags with paragraph separator
		if (dojo.isFF) {
			pasteHtml = pasteHtml.replace(/<br><br>/g, "</p><p>");
		} else if (dojo.isIE) {
			pasteHtml = pasteHtml.replace(/<BR><BR>/g, "</p><p>");
		}
		return pasteHtml;
	},
	
	handleCommentOnPaste: function(pasteHtml, messageHandler, messageBox) {
		// summary:
		//		Perform processing for pasted html in comments.
		// pasteHtml:
		//		Pasted html
		// messageHandler, messageBox:
		//		Used for warning message when pasted content modified
		// returns:
		//		Modified html
		var self = this;
		var pasteDiv = document.createElement("div");
		pasteDiv.innerHTML = pasteHtml;
		
		// Remove embedded artifacts and images from the pasted text.  Comments do not support these.
		var embeds = dojo.query(EditorConstants.RICH_TEXT_EMBED_CONTENT_SELECTOR, pasteDiv);
		var imageErrors = dojo.query(EditorConstants.IMG_EMBED_EDIT_ERROR_CONTENT, pasteDiv);
		var embedWrappers = dojo.query("." + EditorConstants.RICH_TEXT_EMBED_WRAPPER_CLASS, pasteDiv);
		embeds.forEach(function(node) {
			node.parentNode.removeChild(node);
		});
		imageErrors.forEach(function(node) {
			node.parentNode.removeChild(node);
		});
		embedWrappers.forEach(function(node) {
			var embedLinks = dojo.query("a", node);
			// remove the embedded content and replace it with a hyperlink to the artifact
			// instead
			if (embedLinks.length > 0) {
				var embed = embedLinks[0];
				var anchor = node.ownerDocument.createElement("a");
				anchor.setAttribute("href", embed.href);
				anchor.innerHTML = embed.href;
				dojo.place(anchor, node, "before");
			}
			node.parentNode.removeChild(node);
		});
		var images = dojo.query("img", pasteDiv);
		images.forEach(function(node) {
			// remove embedded images, comments do not render embedded images properly
			// TODO - replace image reference with hyperlink to the image, since the
			// wrapped resource is referenced inside rich text, a server query will be
			// necessary to get the associated resource URI for creating the hyperlink
			node.parentNode.removeChild(node);
		});
		var msg = null;
		if (embedWrappers.length > 0) {
			if (images.length > 0) {
				msg = messageHandler.getMessageBoxInfo(7349);
			} else {
				msg = messageHandler.getMessageBoxInfo(7345);
			}
			messageBox.showMessage(msg);
		} else if (images.length > 0) {
			msg = messageHandler.getMessageBoxInfo(7348);
			messageBox.showMessage(msg);
		}
		pasteHtml = pasteDiv.innerHTML;
		dojo.destroy(pasteDiv);
		return pasteHtml;
	},

	getEmbeddedResourceUris: function(rootNode) {
		// summary:
		//		Returns an array of the embedded resource URIs that are found
		//		under rootNode.

		var embeddedResourceUris = [];
		var embeddedContents = dojo.query(EditorConstants.RICH_TEXT_EMBEDS, rootNode);
		embeddedContents.forEach(function(node) {
			embeddedResourceUris.push(node.href);
		});
		return embeddedResourceUris;
	},
	
	removeEmbeddedContentBeforePaste: function(rootNode) {
		// summary:
		//		Removes embedded content, such that embeds are left represented by
		//		only a wrapper node and the link node itself.  The node is marked
		//		with a class, so that the embedded content can be added in again.
		// rootNode:
		//		A node from which to start the replacement.
		// returns:
		//		A count of the number of embeds that were replaced.
		
		// look up the embedded content separately from the wrappers, because on IE they can sometimes
		// be moved from underneath the wrapper node (to be a sibling of it) before we are given the
		// paste HTML by the editor.
		var embeddedContents = dojo.query(EditorConstants.RICH_TEXT_EMBED_CONTENT_SELECTOR, rootNode);
		var embedWrappers = dojo.query("." + EditorConstants.RICH_TEXT_EMBED_WRAPPER_CLASS, rootNode);
		var retCount = embedWrappers.length;
		embeddedContents.forEach(function(node) {
			node.parentNode.removeChild(node);
		});
		embedWrappers.forEach(function(node) {
			// add a marker class so that addEmbeddedContentAfterPaste() can find the node.
			dojo.addClass(node, EditorConstants.EMBEDDED_CONTENT_LOAD_REQD_CLASS);
			// Make sure the <a> tag contains some real text, (not completely empty or containing
			// empty elements).  Without some content IE seems to "optimise" out the empty elements.
			var embedLink = dojo.query("a", node)[0];
			if (embedLink) {
				var textContent = (dojo.isIE && dojo.isIE < 9) ? node.innerText : node.textContent;
				textContent = dojo.trim(textContent);
				if (!textContent) {
					embedLink.innerHTML = "&nbsp;";
				}
			}
		});
		return retCount;
	},
	
	addEmbeddedContentAfterPaste: function(editorContentRootNode, parentResourceURI) {
		// summary:
		//		Adds in the embedded content following a paste operation.  Makes use of the
		//		marker class added by removeEmbeddedContentBeforePaste.
		// editorContentRootNode:
		//		The root node of the editor content containing some pasted embeds.
		// parentResourceURI:
		//		The root artifact uri that contains the embedded content.
		var embeddedNodes = dojo.query(EditorConstants.EMBEDDED_CONTENT_LOAD_REQD_SELECTOR, editorContentRootNode);
		
		if (embeddedNodes.length > 0) {
			var index = -1;
			var loadOneEmbed = function() {
				index++;
				if (index < embeddedNodes.length) {
					var curEmbedNode = embeddedNodes[index];
					dojo.removeClass(curEmbedNode, EditorConstants.EMBEDDED_CONTENT_LOAD_REQD_CLASS);
					var embeddedResourceLoader = new EmbeddedResourceLoader(curEmbedNode, {
						parentResourceURI: parentResourceURI,
						remove: true,
						onCreate: loadOneEmbed});
					embeddedResourceLoader.loadEmbeddedResources();
				}
			};
			
			// Start the load sequence.  As each one is loaded it will start
			// a load of the next.
			loadOneEmbed();
		}
	},
		
	representShortcutKeystroke: function(keystroke) {
		// Represent the given keystroke sequence as a string.  In progress -
		// this method will currently not handle all possible keystroke sequences
		var modifierMap = new dojox.collections.Dictionary();
		modifierMap.add(CKEDITOR.ALT, 'Alt'); // $NON-NLS-1$
		modifierMap.add(CKEDITOR.SHIFT, 'Shift'); // $NON-NLS-1$
		modifierMap.add(CKEDITOR.CTRL, 'Ctrl');	// $NON-NLS-1$
		var modifiers = [CKEDITOR.ALT, CKEDITOR.SHIFT, CKEDITOR.CTRL];
		var quotient,
			modifier,
			presentation = [];
		for (var i=0; i<modifiers.length; i++) {
			modifier = modifiers[i];
			quotient = keystroke/modifiers[i];
			if (quotient>1 && quotient<=2) {
				keystroke -= modifier;
				presentation.push(modifierMap.item(modifier));
			}
		}
		presentation.reverse();
		if (keystroke == 32) {
			presentation.push(editorUtilMessages.space);
		} else {
			presentation.push(String.fromCharCode(keystroke));
		}
		return presentation.join('+');
	},
	
	removeNonBreakingSpaceEntities: function (str) {
		return str.replace(/&nbsp;/g, "&#160;");
	},

	validatePastedXHTMLFragment: function(/*CKEDITOR.event*/evt, /*jazz.ui.MessageBox?*/messageBox, /* boolean */ editorHtml) {
		// summary:
		//		Util function that invoked the servers Validation service
		//		upon a CK Editor paste event.
		//		If validation fails, show the appropriate message in the supplied message box.
		//		editorHtml parameter indicates whether or not we are validating paste for artifact editor (vs. rich text comment)
		
		var valid = true;	//valid default is true, unless the errorback is called
		var errorResponse = null,
			fixed;
		if (!evt.data.html) {
		return;
		}
		if (evt.data.programmatic) {
			// we don't need to validate programmatically pasted data, but we do need
			// to prepare it for the server as things besides removing bad data are done
			// during prepareHtmlForServer
			fixed = this.prepareHtmlForServer(evt.data.html, true, editorHtml);
			evt.data.html = fixed;
			return;
		}

		var self = this;
		// 1. We need to remove the CK Editor/browser added stuff, e.g.
		//  _zipIdx attributes and the like, this will fail to validate
		// and will prevent copying within the same document.
		
		// embedded content does not pass server validation, so this call to prepareHtmlForServer
		// removes the embedded content nodes (note that this content is not saved to the artifact, just the
		// embedded <a> tag is saved)
		fixed = this.prepareHtmlForServer(evt.data.html, true, editorHtml);
		var deferred = ValidationServiceClient.validateXhtmlFragment(fixed);
					
		deferred.addErrback(function (error){
			//if there is an error pasting with html
			
			//TODO: Server throws a 500 when markup is not well formed, it shouldn't do this.
			// in the meantime - hide the messageBox it will put up.
			if (messageBox) {
				MessageHandler._getMessageDisplayWidget().hideMessage();
				var msg = MessageHandler.getMessageBoxInfo(7341);
				messageBox.showMessage(msg);
			}
			
			valid = false;
			errorResponse = error;
		});
		
		if (valid) {
			return null;
		} else {
			return errorResponse;
		}
	},
			
	showInvalidPasteDialog: function(/*CKEDITOR.event*/evt, validationError /* error response  */) {
		// summary:
		//		This will show a dialog asking informing the user that the contents
		//		attempted to paste are invalid, and asking if they would like to paste
		//		as plain text instead. If they press "yes" to pasting as plain, the
		//		pasteAsPlain command in the editor is invoked.
		
		// if user says yes to plaintext paste, open the dialog for plain text paste
		var onOk = function() {
			evt.editor.execCommand('pastetext');
		};
		
		// if user says no to paste as plain text,
		var onCancel = function() {
		
		};
		
		var failedValidationDialogContent = document.createElement("div");
		var dialogMsg = document.createElement("p");
		var errorDetailsMessageBox = new MessageBox();
		
		failedValidationDialogContent.appendChild(errorDetailsMessageBox.domNode);
		failedValidationDialogContent.appendChild(dialogMsg);
		dialogMsg.innerHTML = editorUtilMessages.pasteAsPlainInsteadDialogMessage;
		// TODO - parse the error message so it's easier to read
		var detail = validationError.message;
		var msgArgs = MessageHandler.getMessageBoxInfo(7344, null, detail);
		errorDetailsMessageBox.showMessage(msgArgs);
		
		dojo.style(dialogMsg, {"paddingTop": "10px"});
						
		var invalidClipboardDataDialog = new MessageForm(
			{
				confirmContent: failedValidationDialogContent,
				dialogTitle: editorUtilMessages.pasteAsPlainInsteadDialogTitle,
				dialogStyle: MessageForm.DIALOG_TYPES.YesNo,
				onOk: onOk,
				dialogWidth: "370px",
				onCancel: onCancel
			}
		);
			
		invalidClipboardDataDialog.show();
		
		//clear the paste data, in effect pasting nothing while waiting for user input
		evt.data.html = "";
	},
	
	// Did not want to mess with the SendEmailForm.js
	// Made validatePastedXHTMLFragmentReturnFixed and showInvalidPasteDialogWithFixed for testing
	validatePastedXHTMLFragmentReturnFixed: function(/*CKEDITOR.event*/evt, /*jazz.ui.MessageBox?*/messageBox, /* boolean */ editorHtml) {
		// summary:
		//		Util function that invoked the servers Validation service
		//		upon a CK Editor paste event.
		//		If validation fails, show the appropriate message in the supplied message box.
		//		editorHtml parameter indicates whether or not we are validating paste for artifact editor (vs. rich text comment)
		
		var valid = true;	//valid default is true, unless the errorback is called
		var errorResponse = null,
			fixed;
		if (evt.data.type && evt.data.type != "html") {
		return;
		}
		if (evt.data.programmatic) {
			// we don't need to validate programmatically pasted data, but we do need
			// to prepare it for the server as things besides removing bad data are done
			// during prepareHtmlForServer
			fixed = this.prepareHtmlForServer(evt.data.dataValue, true, editorHtml);
			evt.data.dataValue = fixed;
			return;
		}

		var self = this;
		// 1. We need to remove the CK Editor/browser added stuff, e.g.
		//  _zipIdx attributes and the like, this will fail to validate
		// and will prevent copying within the same document.
		
		// embedded content does not pass server validation, so this call to prepareHtmlForServer
		// removes the embedded content nodes (note that this content is not saved to the artifact, just the
		// embedded <a> tag is saved)
		fixed = this.prepareHtmlForServer(evt.data.dataValue, true, editorHtml);
		var deferred = ValidationServiceClient.validateXhtmlFragment(fixed);
					
		deferred.addErrback(function (error){
			//if there is an error pasting with html
			//TODO: Server throws a 500 when markup is not well formed, it shouldn't do this.
			// in the meantime - hide the messageBox it will put up.
			if (messageBox) {
				MessageHandler._getMessageDisplayWidget().hideMessage();
				var msg = MessageHandler.getMessageBoxInfo(7341);
				messageBox.showMessage(msg);
			}
			
			valid = false;
			errorResponse = error;
		});
		
		if (valid) {
			return null;
		} else {
			return {"error" : errorResponse, "fixed" : evt.data.dataValue} ;
		}
	},
	
	showInvalidPasteDialogWithFixed: function(/*CKEDITOR.event*/evt, validationError /* error response  */, fixed) {
		var that = this;
		// summary:
		//		This will show a dialog asking informing the user that the contents
		//		attempted to paste are invalid, and asking if they would like to paste
		//		as plain text instead. If they press "yes" to pasting as plain, the
		//		pasteAsPlain command in the editor is invoked.
		
		// if user says yes to plaintext paste, open the dialog for plain text paste
		var onOk = function() {
			evt.editor.execCommand('pastetext');
		};
		
		// if user says no to paste as plain text,
		var onNo = function() {
		
		};
		
		var onCancel = function() {
			
		var dialog = new com.ibm.rdm.web.editor.RMCleanDialog({
			sourceHTML: fixed,
			editor: evt.editor
		});
		
		dialog.show();
		
			
		
		};

		
		var failedValidationDialogContent = document.createElement("div");
		var dialogMsg = document.createElement("p");
		var errorDetailsMessageBox = new MessageBox();
		
		failedValidationDialogContent.appendChild(errorDetailsMessageBox.domNode);
		failedValidationDialogContent.appendChild(dialogMsg);
		dialogMsg.innerHTML = editorUtilMessages.pasteAsPlainInsteadDialogMessage;
		// TODO - parse the error message so it's easier to read
		var detail = validationError.message;
		var msgArgs = MessageHandler.getMessageBoxInfo(7344, null, detail);
		errorDetailsMessageBox.showMessage(msgArgs);
		
		dojo.style(dialogMsg, {"paddingTop": "10px"});
						
		var invalidClipboardDataDialog = new MessageForm(
			{
				confirmContent: failedValidationDialogContent,
				dialogTitle: editorUtilMessages.pasteAsPlainInsteadDialogTitle,
				dialogStyle: MessageForm.DIALOG_TYPES.YesNoCancel,
				yesButtonText: editorUtilMessages.pasteAsPlainText,
				noButtonText: editorUtilMessages.doNotPaste,
				cancelButtonText: editorUtilMessages.attemptRM,
				onOk: onOk,
				dialogWidth: "370px",
				onCancel: onCancel,
				onNo: onNo
			}
		);
			
		invalidClipboardDataDialog.show();
		
		//clear the paste data, in effect pasting nothing while waiting for user input
		evt.data.dataValue = "";
	},
	
	
	_getDeepFirstChild: function(cknode) {
		// summary:
		//		Helper to walk down the tree of first cknode children, returning when
		//		down to the full depth (the leftmost leaf in the subtree defined by the
		//		supplied cknode).
		// cknode:
		//		A cknode to start from.
		// returns:
		//		A leaf cknode, which may be the supplied cknode if it has no children.
		while (cknode.type == CKEDITOR.NODE_ELEMENT && cknode.getChildCount() > 0) {
			cknode = cknode.getFirst();
		}
		return cknode;
	},

	
	_getDeepLastChild: function(cknode) {
		// summary:
		//		Helper to walk down the tree of last cknode children, returning when
		//		down to the full depth (the rightmost leaf in the subtree defined by the
		//		supplied cknode).
		// cknode:
		//		A cknode to start from.
		// returns:
		//		A leaf cknode, which may be the supplied cknode if it has no children.
		while (cknode.type == CKEDITOR.NODE_ELEMENT && cknode.getChildCount() > 0) {
			cknode = cknode.getLast();
		}
		return cknode;
	},


	_getNextLeafInBlock: function(cknode) {
		// summary:
		//		Helper to find the next leaf cknode in the tree (i.e. the first leaf cknode encountered
		//		in an rightwards walk of the tree). The search is restricted to the block element
		//		containing the given cknode.
		// cknode:
		//		A cknode to start the search from. Must be a leaf cknode (but not necessarily a text cknode).
		// returns:
		//		A leaf cknode, or null if no leaf is found in the block.
		var next;
		while (true) {
			if (cknode.hasNext()) {
				if(cknode.$.nextSibling.nodeName !="BR"){
					next = cknode.getNext();		
					if (!this._isInlineNode(next)) {
						return null;
					}
					break;
				}
			}
			cknode = cknode.getParent();
			if (!this._isInlineNode(cknode)) {
				return null;
			}
		}
		next = this._getDeepFirstChild(next);
		return next;
	},
	
	
	_getPreviousLeafInBlock: function(cknode) {
		// summary:
		//		Helper to find the previous leaf cknode in the tree (i.e. the first leaf cknode encountered
		//		in an leftwards walk of the tree). The search is restricted to the block element
		//		containing the given cknode.
		// cknode:
		//		A cknode to start the search from. Must be a leaf cknode (but not necessarily a text cknode).
		// returns:
		//		A leaf cknode, or null if no leaf is found in the block.
		var previous;
		while (true) {
			if (cknode.hasPrevious()) {
				if(cknode.$.previousSibling.nodeName !="BR"){
					previous = cknode.getPrevious();
					if (!this._isInlineNode(previous)) {
						return null;
					}
					break;
				}
			}
			cknode = cknode.getParent();
			if (!this._isInlineNode(cknode)) {
				return null;
			}
		}
		previous = this._getDeepLastChild(previous);
		return previous;
	},
	
		
	getNextTextNodeInBlock: function(cknode) {
		// summary:
		//		Gets the next text cknode in the tree (i.e. the first text cknode encountered
		//		in a rightwards walk of the tree).  The search is restricted to the block element
		//		containing the given cknode.
		// cknode:
		//		A cknode to start the search from. Must be a leaf cknode (but not necessarily a text cknode).
		// returns:
		//		A text cknode, or null if no text is found in the block.
		while (true) {
			cknode = this._getNextLeafInBlock(cknode);
			if (!cknode) {
				break;
			}
			if (cknode.type == CKEDITOR.NODE_TEXT) {
				break;
			}
		}
		return cknode;
	},
	
	getPreviousTextNodeInBlock: function(cknode) {
		// summary:
		//		Gets the previous text cknode in the tree (i.e. the first text cknode encountered
		//		in a leftwards walk of the tree).  The search is restricted to the block element
		//		containing the given cknode.
		// cknode:
		//		A cknode to start the search from. Must be a leaf cknode (but not necessarily a text cknode).
		// returns:
		//		A text cknode, or null if no text is found in the block.
		while (true) {
			cknode = this._getPreviousLeafInBlock(cknode);
			if (!cknode) {
				break;
			}
			if (cknode.type == CKEDITOR.NODE_TEXT) {
				break;
			}
		}
		return cknode;
	},
	
	_isInlineNode: function(cknode) {
		// returns:
		//		true if the given cknode is an inline element (rather than a block element).
		if (cknode.type == CKEDITOR.NODE_TEXT) {
			return true;
		}
		var displayStyle = dojo.style(cknode.$, 'display');
		return displayStyle == 'inline';
	},
	
	getWordRangeForSelection: function(selection) {
		// summary:
		//		Builds a range that spans a single word.
		// selection:
		//		A selection.  It is assumed that this selection contains a single range, and
		//		that it is collapsed (i.e. it has no content - it just indicates the cursor position).
		// returns:
		//		A hash of the form {range: range, word: string}
		//		The range encompasses a word that preceeds or surrounds the selection.  The range does not
		//		span into another block so will not bleed from one paragraph to the next, but may include
		//		inline elements (<span> <b> <i> etc.)
		//		Word is the text included in range.
		//		null is returned if no word can be found.
		
		var ranges = selection.getRanges();
		var range = ranges[0];
		var startNode = range.startContainer;
		var index;
		var text;
		var workingNode;
		var curChar;
		var selectedWord;
		
		// Make sure we are starting on a text node

		if (startNode.type == CKEDITOR.NODE_TEXT) {
			text = startNode.getText();
			index = range.startOffset;
		} else {
			// we have an element - start searching from the child indicated by offset in the range.
			// Workaround - If we are right at the end of a block, startOffset can be (one) greater than the
			// greatest index of a child node, so limit if necessary.  And in this case, make sure we start the
			// text search from the end of the available text.
			var childCount = startNode.getChildCount();
			if (childCount == 0) {
				// best we can do is start at the node we've got
				text = startNode.getText();
				index = 0;
			}
			else if (range.startOffset >= childCount) {
				// probably at the end of a block.
				// make sure we start the text search from the rightmost thing we have.
				startNode = this._getDeepLastChild(startNode);
				text = startNode.getText();
				index = text.length - 1;
			} else {
				// the usual case - we have a start offset that indicates a child, so start from there.
				startNode = startNode.getChild(range.startOffset);
				startNode = this._getDeepFirstChild(startNode);
				text = startNode.getText();
				index = 0;
			}
		}
		workingNode = startNode;

		
		// scan backwards over any whitespace to find the last character in the word if we
		// are beyond the end of it.
		var lastWordCharOffset;
		var lastWordCharContainer;
		while (true) {
			if (index < 0) {
				// move to previous text node
				workingNode = this.getPreviousTextNodeInBlock(workingNode);
				if (!workingNode) {
					break;
				}
				text = workingNode.getText();
				index = text.length-1;
			}
			
			curChar = text.charAt(index);
			if (REGEX_WORD_CHARACTERS.test(curChar)) {
				// word character we have reached the end of whitespace
				lastWordCharOffset = index;
				lastWordCharContainer = workingNode;
				break;
			}
			index--;
		}
		
		if (!workingNode) {
			// failed to find a word in the search
			return null;
		}
		
		// scan backwards to find the first character of the word.  We know that this loop will go
		// round at least once, because the index is on a word character found in the previous loop,
		// to the output variables will be set.
		var firstWordCharOffset;
		var firstWordCharContainer;
		selectedWord = "";					// we build up the selected characters in this string.
		while (true) {
			if (index < 0) {
				// move to previous text node
				workingNode = this.getPreviousTextNodeInBlock(workingNode);
				if (!workingNode) {
					break;
				}
				text = workingNode.getText();
				index = text.length-1;
				if (text.length == 0) {
					// no chars in this block - move on to the next.
					continue;
				}
			}
			
			curChar = text.charAt(index);
			if (REGEX_NON_WORD_CHARACTERS.test(curChar)) {
				// non-word character - we have reached the end of the word.
				break;
			}
			firstWordCharOffset = index;
			firstWordCharContainer = workingNode;
			selectedWord = curChar + selectedWord;
			index--;
		}
		
		// scan forwards from the end position we previously found, to include the part of the
		// word to the right of the cursor.
		workingNode = lastWordCharContainer;
		text = workingNode.getText();
		index = lastWordCharOffset+1;
		
		while (true) {
			if (index >= text.length) {
				// move to the next text node
				workingNode = this.getNextTextNodeInBlock(workingNode);
				if (!workingNode) {
					break;
				}
				text = workingNode.getText();
				index = 0;
				if (text.length == 0) {
					// no chars in this block - move on to the next
					index++;
					continue;
				}
			}
			
			curChar = text.charAt(index);
			if (REGEX_NON_WORD_CHARACTERS.test(curChar)) {
				// non-word character - we have reached the end of the word.
				break;
			}
			lastWordCharOffset = index;
			lastWordCharContainer = workingNode;
			selectedWord += curChar;
			index++;
		}
		
		// build a range to return
		var wordRange = new CKEDITOR.dom.range(selection.document);
		wordRange.setStart(firstWordCharContainer, firstWordCharOffset);
		wordRange.setEnd(lastWordCharContainer, lastWordCharOffset+1);
		
		return {range: wordRange, word: selectedWord};
	},
	
	getSelectionPosition: function(editor, selection, selectedRange) {
		// summary:
		//		Gets the position of the start and end of a selection.  Note that this method changes the selection,
		//		and can change the DOM, so any previously discovered ranges should be considered invalid following
		//		execution of this method.
		// editor:
		//		the ckeditor that has a selected range.
		// selection:
		//		The selection, as read from the editor.
		// selectedRange:
		//		A range that has previously been selected within the editor.
		// returns:
		//		A hash of the form {start: startPosition, end: endPosition}, where startPosition and endPosition
		//		are positions as returned from dojo.position();
		var position;
		var domNode;

		// create a bookmark with the required range, which inserts some span elements into the DOM at the start and
		// end of the selection.  We can use the positions of these elements to give the start and end positions
		// of the selection.
		var bookmark = selectedRange.createBookmark();
			
		// Get the positions of the bookmark. Unfortunately they have display=none, so do not
		// have any position information, so make them display briefly to take the positions.
		domNode = bookmark.startNode.$;
		var origDisplay = dojo.style(domNode, "display");
		dojo.style(domNode, "display", "");
		var startPosition = dojo.position(domNode);
		dojo.style(domNode, "display", origDisplay);
		
		domNode = bookmark.endNode.$;
		origDisplay = dojo.style(domNode, "display");
		dojo.style(domNode, "display", "");
		var endPosition = dojo.position(domNode);
		dojo.style(domNode, "display", origDisplay);
		
		// Even though the bookmark spans have display=none, they seem to occupy space in IE, opening up a space
		// on either side of the selected text.  So we remove selecting the bookmark.  This leaves the DOM equivalent
		// to what it was on entry, but it may not be identical, since in some browsers (IE) the text nodes can be left
		// split into separate nodes where the bookmark was.
		selection.selectBookmarks([bookmark]);
		
		// Adjust the positions to account for the positions we've just read being relative to the editor iframe, rather than the
		// overall window.
		
		var iframe = this.getIframeNodeForCKEditor(editor);
		var iframePosition = dojo.position(iframe);
		startPosition.x += iframePosition.x;
		startPosition.y += iframePosition.y;
		endPosition.x += iframePosition.x;
		endPosition.y += iframePosition.y;
		
		position = {start: startPosition, end: endPosition};
		
		return position;
	},
	
	getIframeNodeForCKEditor: function(editor) {
		// summary:
		//		Gets the iframe node that relates to a given ckeditor instance.
		var name = editor.name;
		return dojo.query("#cke_" + name + " iframe")[0];
	},
	
	legaliseEmbeds: function(node) {
		// summary:
		//		Converts embeds located inside ["p", "h1", "h2", "h3", "h4", "h5", "h6"] tags to be nested in div tags. 
		//		This is done for two reasons 
		//		1) to create valid html (<div> elements should not live within <p> tags)
		//		2) Embeded nodes inside <h*> tags cause font inconsistencies. (Imagine <h1>Hello<h1> World</h1></h1>)
		var embeds = dojo.query("."+EditorConstants.RICH_TEXT_EMBED_WRAPPER_CLASS, node);
		embeds.forEach(function(embedNode){
			var ancestorToChange = domutils.findAncestorByTags(embedNode, ["p", "h1", "h2", "h3", "h4", "h5", "h6"], node);
			if (ancestorToChange) {
				// replace the <p> by a <div> with a class that we can use to style it as a <p> and find it later.
				var newDiv = domutils.changeNodeTag(ancestorToChange, "div");
				dojo.addClass(newDiv, EditorConstants.REPLACED_EMBED_PARAGRAPH_ELEMENT_CLASS);
			}
		});

	}
		
};
}());
