/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("com.ibm.rdm.web.review.reviewUpdateMediator");

dojo.require("dojo.i18n");

dojo.require("com.ibm.rdm.web.common.UpdateContext");
dojo.require("com.ibm.rdm.web.common.MessageForm");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.review.ReviewCache");
dojo.require("com.ibm.rdm.web.review.ReviewSaver");
dojo.require("com.ibm.rdm.web.review.ParticipantResultSaver");
dojo.require("com.ibm.rdm.web.review.reviewChangeNotifier");
dojo.require("com.ibm.rdm.web.commenting.ReviewCommentForm");
dojo.require("com.ibm.rdm.web.commenting.CommentingClient");
dojo.require("com.ibm.rdm.web.review.reviewTypes");
//CIRC: dojo.require("com.ibm.rdm.web.review.reviewutils");

dojo.requireLocalization("com.ibm.rdm.web.review", "reviewUpdateMediator");

var MessageForm = com.ibm.rdm.web.common.MessageForm;
var ParticipantResultSaver = com.ibm.rdm.web.review.ParticipantResultSaver;
var ReviewCache = com.ibm.rdm.web.review.ReviewCache;
var ReviewSaver = com.ibm.rdm.web.review.ReviewSaver;
var UpdateContext = com.ibm.rdm.web.common.UpdateContext;
var messages = dojo.i18n.getLocalization("com.ibm.rdm.web.review", "reviewUpdateMediator");
var reviewChangeNotifier = com.ibm.rdm.web.review.reviewChangeNotifier;
var reviewTypes = com.ibm.rdm.web.review.reviewTypes;
//var reviewutils = com.ibm.rdm.web.review.reviewutils;


com.ibm.rdm.web.review.reviewUpdateMediator = {
	// summary:
	//		This singleton serves as a mediator to coordinate the work that goes
	//		when attempting to update review related information.
	// description:
	//		Reviews are actually composed of "review" artifacts and "participant
	//		result" artifacts.  Whenever the user attempts to persist review
	//		related information, this class is responsible for making it happen.
	//		For example if the user performs vote casts with an ApprovalEditor
	//		or with an ArtifactInReviewViewer, this class is called to make
	//		sure that the appropriate POST or PUT of a "participant result"
	//		occurs.  As another example, if the user attempts to "Continue
	//		Review" after adding some new artifacts, this class is called to
	//		make sure that an appropriate PUT for the "review" artifact occurs.

	///////////////////////////////////////////////////////////////////////////
	// Private fields
	
	///////////////////////////////////////////////////////////////////////////
	// API

	castArtifactVotes: function(
			reviewParser, myChangedArtifactInfos, doPromptForComments,
			onSaveStarted,
			onSaveComplete, onSaveCancelled, onSaveError) {
		// summary:
		//		Save away new artifact vote casts for the the current user.
		// reviewParser:
		//		The ReviewParser instance that was originally used to hold
		//		the pre-tweaked myChangedArtifactInfos (i.e. before votes were
		//		"cast" within them by changing their myArtifactStatusType
		//		property field value).
		// myChangedArtifactInfos:
		//		Please note that the artifactInfos found in this array represent
		//		ONLY those artifactInfo's whose vote setting has changed (i.e.
		//		it is NOT intended to represent the complete list of new values
		//		to record [changed vote values as well as original vote values]).
		// doPromptForComments:
		//		Only prompt for comments when this value is set to true.
		// onSaveStarted:
		//		A callback that will be invoked when the save has started.
		// onSaveComplete:
		//		A callback that will be invoked when the save has successfully
		//		completed.
		// onSaveCancelled:
		//		A callback that will be invoked when the user decides to
		//		Cancel the vote cast he/she was about to make.
		// onSaveError:
		//		A callback that will be invoked when the POST or PUT on the
		//		participant results fails.  For example, perhaps the review
		//		has been deleted (error code 410).  Callers should be aware
		//		that when errors occur, the ErrorSelectionWidget shows the
		//		error BEFORE is callback is invoked.
		if (doPromptForComments) {
			var onDialogClose = function() {
			};
			var onCancel = function() {
				onSaveCancelled();
			};
			// Note: In the following form, when performCastArtifactVotesAndAddComments
			// is called we're explicitly passing skipErrorsWhenSavingParticipantResults
			// as 'true' so that the form code can receive and deal with an abnormal
			// return status (as we want to address the error in the form dialog rather
			// than at the page level [at the top of the page where we normally show
			// error messages] to enable the user to copy the comment text to the
			// clipboard).
			var form = com.ibm.rdm.web.commenting.ReviewCommentForm.getInstance({
				artifactStatus: myChangedArtifactInfos[0].myArtifactStatusType,
				isMultipleArtifacts: (myChangedArtifactInfos.length > 1),
				componentURI: reviewParser.getComponentUri(),
				formData: {},
				onDialogClose: onDialogClose,
				onOk: onSaveComplete,
				onCancel: onCancel,
				performCastArtifactVotesAndAddComments: dojo.hitch(
					this,
					function(onSaveComplete, onSaveError){
						onSaveStarted();
						this._castArtifactVotesAndAddComments(
							reviewParser, myChangedArtifactInfos,
							onSaveComplete, onSaveCancelled, onSaveError,
							form.commentText,
							true // skipErrorsWhenSavingParticipantResults
						); 
					}
				)
			}); 
			form.show();
		} else {
			onSaveStarted();
			this._castArtifactVotesAndAddComments(
				reviewParser, myChangedArtifactInfos,
				onSaveComplete, onSaveCancelled, onSaveError, "");
		}
	},

	markMyReviewStatusAsDone: function(
			reviewParser, onSaveStarted,
			onSaveComplete, onSaveCancelled, onSaveError) {
		// summary:
		//		Mark the current user's reviewing state as "Done" for the
		//		entire review.
		// reviewParser:
		//		A ReviewParser instance.
		// onSaveStarted:
		//		A callback that will be invoked when the save has started.
		// onSaveComplete:
		//		A callback that will be invoked when the save has successfully
		//		completed.
		// onSaveCancelled:
		//		A callback that will be invoked when the save has been cancelled.
		// onSaveError:
		//		A callback that will be invoked when the save has failed.
		var self = this;
		// Get confirmation from the user if artifacts are in progress or not started.
		// Only continue with the save if confirmation received.
		var artifactInfos = reviewParser.getArtifactInfos();
		var abstainedArtifactInfos = [];
		for (var i = 0; i < artifactInfos.length; i++) {
			var artifactInfo = artifactInfos[i];
			if (artifactInfo.myArtifactStatusType === reviewTypes.ARTIFACT_STATUS_TYPES.NotStarted ||
				artifactInfo.myArtifactStatusType === reviewTypes.ARTIFACT_STATUS_TYPES.InProgress) {
				abstainedArtifactInfos.push(artifactInfo);
			}
		}
		var message = null;
		if (abstainedArtifactInfos.length === 0) {
			message = messages["confirmDoneAllVotesAlreadyCastMessage"];
		}
		else if (abstainedArtifactInfos.length === 1) {
			message = messages["confirmDoneSingleArtifactMessage"];
		}
		else {
			message =
				dojo.string.substitute(
					messages["confirmDoneMultipleArtifactsMessage"],
					[abstainedArtifactInfos.length]
				);
		}
		var onDialogClose = function() {
			onSaveCancelled();
		};
		var doDone = function() {
			onSaveStarted();
			self._markMyReviewStatusAsDone(
				reviewParser, onSaveComplete, onSaveError, abstainedArtifactInfos);
		};
		var doCancel = function() {
			onSaveCancelled();
		};
		var form = new MessageForm(
			{
				confirmMessage: message,
				dialogTitle: messages["confirmDone"],
				dialogStyle: MessageForm.DIALOG_TYPES.YesNo,
				onDialogClose: onDialogClose,
				onOk: doDone,
				onCancel: doCancel
			}
		);   
		form.show();
	},

	artifactUnderReviewHasBeenOpened: function(reviewParser, artifactInfo) {
		// summary:
		//		If appropriate, change the state of the artifact under review
		//		from "Not Started" to "In Progress".
		// reviewParser:
		//		A ReviewParser instance.
		// artifactInfo:
		//		Represents the artifact that has just been opened.
		// description:
		//		The first time a review participant opens an artifact under
		//		review from a Started review, a state transition is supposed to
		//		occur for the user's artifact result.  If the artifact's current
		//		status is "Not Started" we move it to "In Progress", otherwise
		//		we do nothing.
		com.ibm.rdm.web.review.reviewutils.canVote(reviewParser).addCallback(this, function(canVote){
			if (canVote
				&& reviewParser.getReviewState() === reviewTypes.REVIEW_STATE_TYPES.Started
				&& artifactInfo.myArtifactStatusType === reviewTypes.ARTIFACT_STATUS_TYPES.NotStarted
				) {
				artifactInfo.myArtifactStatusType = reviewTypes.ARTIFACT_STATUS_TYPES.InProgress;
				var self = this;
				var onParticipantResultSaverSaveHasCompleted = function() {
					self._updateReviewCacheAndSendChangeNotifications(reviewParser, true);
				};
				var participantResultSaver = new ParticipantResultSaver({
					reviewParser: reviewParser
				});
				participantResultSaver.castArtifactVotes(
					[ artifactInfo ],
					onParticipantResultSaverSaveHasCompleted
				);
			} else {
				// Send out change notifications.  We do this because opening a new
				// ArtifactInReviewViewer results in getting the most up to date
				// information about the Review it's bound to (represented by
				// reviewParser) and this information MAY contain updates to the
				// information sent out the last time the notifications were sent
				// out for this Review.  For example, perhaps the Review state has
				// changed from Started to Paused, in which case appropriate UI
				// updates would need to be performed in the viewers bound to this
				// Review.
				reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
			}			
			return canVote;
		});

	},

	reviewHasBeenOpened: function(reviewParser) {
		// summary:
		//		If the Review that's just been opened has a more recent etag than
		//		the last one loaded by the ReviewCache, we need to have the
		//		ReviewCache reload it and send out change notifications.
		// reviewParser:
		//		A ReviewParser instance.
		var lastLoadedETag =
			ReviewCache.getETagLastLoadedForReview(reviewParser.getSelfUri());
		if (lastLoadedETag && (lastLoadedETag !== reviewParser.getETag())) {
			this._updateReviewCacheAndSendChangeNotifications(reviewParser);
		}
	},

	saveReview: function(reviewParser,
						 /*Callback?*/onSaveComplete,
						 /*Callback?*/onSaveError,
						 /*Boolean?*/doAvoidSendingChangeNotification) {
		// summary:
		//		Save a new version of the review.
		// description:
		//		Either the review is in the Draft, Paused or Reviewed state and
		//		edits were made and Save was called OR the user clicked on a
		//		state transition button (e.g. it was in "Draft" state and he/she
		//		clicked on "Start Review").  In either case the appropriate
		//		changes are already reflected in the passed 'reviewParser' and
		//		we need to save these changes.
		// reviewParser:
		//		A ReviewParser instance.
		// onSaveComplete:
		//		A callback that will be invoked when the save has successfully
		//		completed.
		// onSaveError:
		//		A callback that will be invoked when the save has failed.
		// doAvoidSendingChangeNotification:
		//		This param should be passed as true if you know that right after
		//		saving, you're immediately going to perform a refresh of the
		//		review (i.e. you wish to defer updates until AFTER this
		//		subsequent refresh has been performed).
		var reviewSaver = new ReviewSaver({
			reviewParser: reviewParser
		});
		// FIXME Add error handling case
		var onReviewSaverSaveHasCompleted = function(response) {
			if (!doAvoidSendingChangeNotification) {
				reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
			}
			if (onSaveComplete) {
				onSaveComplete(response);
			}
		};
		return reviewSaver.saveReview(onReviewSaverSaveHasCompleted,
			onSaveError);
	},

	createNewReview: function(reviewParser,
							  /*Callback?*/onSaveComplete,
							  /*Callback?*/onSaveError) {
		// summary:
		//		The user has just clicked on "Finish" in a "New Review" wizard.
		//		This function is called to persist the initial version of a
		//		newly created review.
		// reviewParser:
		//		A ReviewParser instance.
		// onSaveComplete:
		//		A callback that will be invoked when the save has successfully
		//		completed.
		// onSaveError:
		//		A callback that will be invoked when the save has failed.
		var reviewSaver = new ReviewSaver({
			reviewParser: reviewParser
		});
		reviewSaver.saveReview(onSaveComplete, onSaveError);
	},

	saveParticipantVotes: function(reviewParser,
								   userURI,
								   isDone,
								   onSaveComplete,
								   onSaveError) {
		// summary:
		//		This function is called to perform bulk changes for the votes
		//		cast by a single participant.
		// reviewParser:
		//		A ReviewParser instance.
		// userURI:
		//		The userURI associated with the participant that we're
		//		targetting for resetting his/her "doneness".
		// isDone:
		//		A boolean representing whether the participant result to be
		//		updated for participant 'userURI' should record this
		//		participant as being done or not yet done.
		// artifactStatusType:
		//		When isDone is passed as true, this parameter holds the
		//		artifact status type that should be recorded for all remaining
		//		artifacts (i.e. those artifacts that this participant has not
		//		yet voted on).
		var self = this;
		var participantInfos = reviewParser.getParticipantInfos();
		var participantInfo = participantInfos[userURI];
		var participantResultSaver = new ParticipantResultSaver({
			reviewParser: reviewParser
		});
		var onSaveParticipantResultComplete = function() {
			onSaveComplete();
		};
		participantResultSaver.resetArtifactVotes(
			participantInfo,
			isDone,
			null,
			[],
			onSaveParticipantResultComplete,
			onSaveError
		);
	},
	
	getSaveParticipantVotesRequestArgs: function(reviewParser, userURI,
		isDone) {
		// summary:
		//		Generate the multi-request arguments for saving all
		//		participant votes 
		// reviewParser:
		//		A ReviewParser instance.
		// userURI:
		//		The userURI associated with the participant that we're
		//		targetting for resetting his/her "doneness".
		// isDone:
		//		A boolean representing whether the participant result to be
		//		updated for participant 'userURI' should record this
		//		participant as being done or not yet done.
		var participantInfos = reviewParser.getParticipantInfos();
		var participantInfo = participantInfos[userURI];
		var participantResultSaver = new ParticipantResultSaver({
			reviewParser: reviewParser
		});
		return participantResultSaver.getSaveArtifactVotesArgs(
			participantInfo, isDone, null, []);
	},

	handleRefresh: function(reviewParser, /*Callback?*/onRefreshComplete,
		/*Boolean?*/thisIsTheNewReviewParserDontGetItAgain) {

		if(!thisIsTheNewReviewParserDontGetItAgain) {
			var onReviewLoaded = function(response) {
				var reviewParser = response.getReviewParser();
				reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
				if (onRefreshComplete) {
					onRefreshComplete();
				}
				return response;
			};
			var deferred = ReviewCache.getReview(reviewParser.getReviewId());
			deferred.addCallback(onReviewLoaded);
		}
		else {
			reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
			if(onRefreshComplete) {
				onRefreshComplete();
			}
		}
	},

	reviewSaveInProgress: function(reviewParser) {
		// summary:
		//		A review or the user's participant results is about to be saved
		//		to the server.
		// description:
		//		Make sure that registered observers are told that the review is
		//		about to be saved.
		// reviewParser:
		//		A ReviewParser object that contains the last loaded review
		//		information as well as the "working copy" updates that the current
		//		user is attempting to save (for example, a review state transition
		//		from "Started" to "Paused" or a vote casting change from "Not
		//		Started" to "Reviewed").
		reviewChangeNotifier.reviewSaveIsInProgress(reviewParser);
	},

	reviewSaveHasCompleted: function(reviewParser,
									 etagOfReviewOrParticipantResult,
									 /*String?*/participantResultUri,
									 /*String?*/reviewRevision) {
		// summary:
		//		This function is called whenever a "save review" or
		//		"save participant results" (but ONLY for the current
		//		user) request has succeeded.
		// description:
		//		An attempt at saving the review or the current user's
		//		participant results has just succeeded.  Fold the new state
		//		information found in the 'participantResultUri' and
		//		'participantResultUri' arguments into the 'reviewParser'
		//		and forward this updated 'reviewParser' to the appropriate
		//		function in the reviewChangeNotifier (so that registered
		//		observers are told that the review or participant result
		//		has just been successfully saved).
		// reviewParser:
		//		A ReviewParser object that contains the last loaded review as
		//		well as vote cast changes that the current user has just
		//		succeeded in saving.
		// etagOfReviewOrParticipantResult:
		//		If participantResultUri is passed, this represents the new
		//		ETag associated with the newly saved participant result.
		//		Otherwise, if participantResultUri is NOT passed, this
		//		represents the new review ETag.
		// participantResultUri:
		//		The participant results URI that was just successfully saved.
		//		This argument is NOT passed if the user is just attempting to
		//		save a review (as opposed to casting a vote).
		if (participantResultUri) {
			reviewParser.setParticipantResultUri(participantResultUri);
			reviewParser.setParticipantResultETag(
				etagOfReviewOrParticipantResult);
		} else {
			reviewParser.setRevisionAndETag(reviewRevision, etagOfReviewOrParticipantResult);
		}
		reviewChangeNotifier.reviewSaveHasCompleted(reviewParser);
	},

	reviewSaveHasFailed: function(reviewParser, errorStatus) {
		// summary:
		//		This function is called whenever a "save participant results"
		//		request has returned with an error code.
		// description:
		//		Make sure that registered observers are told that the review save
		//		request failed.
		// reviewParser:
		//		A ReviewParser object that contains the last loaded review
		//		information as well as the "working copy" updates that the current
		//		user failed in attempting to save (for example, a review state
		//		transition from "Started" to "Paused" or a vote casting change
		//		from "Not Started" to "Reviewed").
		reviewChangeNotifier.reviewSaveHasFailed(reviewParser, errorStatus);
	},

	///////////////////////////////////////////////////////////////////////////////////
	// Helpers

	_updateReviewCacheAndSendChangeNotifications: function(reviewParser, /*Boolean?*/doExpectVotesFromMe) {
		// Note: We need to make sure to change the UpdateContext so that
		// the ReviewCache.getReview function will know that it needs to
		// bother querying the server for the new version of the Review
		// artifact.
		UpdateContext.cachedArtifactHasBeenSaved();
		var onReviewLoaded = function(response) {
			var reviewParser = response.getReviewParser();
			reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
			return response;
		};
		var deferred = ReviewCache.getReview(reviewParser.getReviewId(), doExpectVotesFromMe);
		deferred.addCallback(onReviewLoaded);
	},

	_castArtifactVotesAndAddComments: function(
		reviewParser, 
		myChangedArtifactInfos,
		onSaveComplete,
		onSaveCancelled,
		onSaveError,
		commentText,
		/*Boolean?*/skipErrorsWhenSavingParticipantResults
	) {
		var self = this;
		var participantResultSaver = new ParticipantResultSaver({
			reviewParser: reviewParser
		});
		var onParticipantResultSaverSaveHasCompleted = function() {
			// Note: We need to make sure to change the UpdateContext so that
			// the ReviewCache.getReview function will know that it needs
			// bother querying the server for the new version of the Review
			// artifact.
			UpdateContext.cachedArtifactHasBeenSaved();
			var onReviewLoaded = function(response) {
				var reviewParser = response.getReviewParser();
				onSaveComplete();
				reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
				self._createComments(
					reviewParser, myChangedArtifactInfos, commentText);
				return response;
			};
			var deferred = ReviewCache.getReview(
				reviewParser.getReviewId(),
				true			// doExpectVotesFromMe
			);
			deferred.addCallback(onReviewLoaded);
		};
		participantResultSaver.castArtifactVotes(
			myChangedArtifactInfos,
			onParticipantResultSaverSaveHasCompleted,
			onSaveError,
			skipErrorsWhenSavingParticipantResults
		);
	},
	
	_createComments: function(reviewParser, myChangedArtifactInfos, commentText) {
		var self = this;
		var commentingWidgets = dijit.registry.byClass("com.ibm.rdm.web.commenting.CommentingWidget");
		var commentAddedCount = 0;
		var commentingEnabled = (com.ibm.rdm.web.FRONTING_PROPERTY_REVIEW_VOTING_COMMENTS_ENABLED === "true");
		if(commentingEnabled || dojo.trim(commentText) !== "") {
			var commentsToAdd = [];
			
			var onCommentsAdded = function(response) {
				// All the comments have been created, update the review's artifact comments widget to reflect
				// the new comments.
				var reviewCommentsWidget = null;
				var reviewCommentsWidgets = dijit.registry.byClass("com.ibm.rdm.web.review.ReviewArtifactsCommentWidget");
				reviewCommentsWidgets.forEach(function(w){
					if (w.resourceURI === reviewParser.getSelfUri()) {
						reviewCommentsWidget = w;
					}
				});
				if (reviewCommentsWidget) {
					reviewCommentsWidget.refreshComments();
				}
				return response;
			};
			
			dojo.forEach(myChangedArtifactInfos, function(artifactInfo) {
				var subject = null;
				var resourceURI = artifactInfo.resourceURI;
				var priority = com.ibm.rdm.web.commenting.PRIORITY_DEFAULT;
				switch (artifactInfo.myArtifactStatusType) {
					case reviewTypes.ARTIFACT_STATUS_TYPES.Approved:				
						subject = dojo.string.substitute(messages.approvedSubject, [reviewParser.getReviewName()]);
						break;
					case reviewTypes.ARTIFACT_STATUS_TYPES.Disapproved:
						subject = dojo.string.substitute(messages.disapprovedSubject, [reviewParser.getReviewName()]);
						priority = com.ibm.rdm.web.commenting.PRIORITY_HIGH;
						break;
					case reviewTypes.ARTIFACT_STATUS_TYPES.Abstained:
						subject = dojo.string.substitute(messages.abstainedSubject, [reviewParser.getReviewName()]);
						break;
					case reviewTypes.ARTIFACT_STATUS_TYPES.Reviewed:
						subject = dojo.string.substitute(messages.reviewedSubject, [reviewParser.getReviewName()]);
						break;
					case reviewTypes.ARTIFACT_STATUS_TYPES.InProgress:
						subject = dojo.string.substitute(messages.abstainedSubject, [reviewParser.getReviewName()]);
						break;
					case reviewTypes.ARTIFACT_STATUS_TYPES.NotStarted:
						subject = dojo.string.substitute(messages.abstainedSubject, [reviewParser.getReviewName()]);
						break;
				}
				if (subject === null) { 
					var details = "function reviewUpdateMediator._createComments"; // NON-NLS
					com.ibm.rdm.web.common.MessageHandler.showError(7306,
						[artifactInfo.myArtifactStatusType], details);
				} else {
					var newComment = new com.ibm.rdm.web.commenting.Comment({
						locationId: "", // artifact level comment 
						reviewId: reviewParser.getReviewId(),
						text: commentText,
						title: subject,
						priority: priority, 
						resourceUri: resourceURI,
						parentComment: "",
						rootComment: "",
						directedToUsers: [],
						isResolved: false,
						isRoot: true
					});
					commentsToAdd.push(newComment);
				 }
			});

			var addCommentsDfd = com.ibm.rdm.web.commenting.CommentingClient
				.saveNewComments(commentsToAdd, reviewParser.getComponentUri());

			addCommentsDfd.addCallback(this, onCommentsAdded);
			addCommentsDfd.addErrback(this, function(error) {
				com.ibm.rdm.web.common.MessageHandler.showError(7307, null,
					error);
			});
		}
	},
	
	_markMyReviewStatusAsDone: function(reviewParser, onSaveComplete,
										onSaveError, abstainedArtifactInfos) {
		var self = this;
		var participantResultSaver = new ParticipantResultSaver({
			reviewParser: reviewParser
		});
		var onParticipantResultSaverSaveHasCompleted = function() {
			// Note: We need to make sure to change the UpdateContext so that
			// the ReviewCache.getReview function will know that it needs
			// bother querying the server for the new version of the Review
			// artifact.
			UpdateContext.cachedArtifactHasBeenSaved();
			self._createComments(reviewParser, abstainedArtifactInfos, "");
			var onReviewLoaded = function(response) {
				var reviewParser = response.getReviewParser();
				// We just obtained a new instance of a ReviewParser which
				// accurately reflects the newly marked "doneness" (and
				// possibly new vote casts for promoting any "Not Started"
				// artifacts for this review to be "Abstained").
				// Now we need to inform all listeners so that the appropriate
				// UI updates may occur.
				onSaveComplete();
				reviewChangeNotifier.reviewHasBeenUpdated(reviewParser);
				return response;
			};
			var deferred = ReviewCache.getReview(
				reviewParser.getReviewId(),
				true			// doExpectVotesFromMe
			);
			deferred.addCallback(onReviewLoaded);
		};
		participantResultSaver.markMyReviewStatusAsDone(
			onParticipantResultSaverSaveHasCompleted, onSaveError);
	}
};

})();
